@@ -1,5 +1,5 @@
 #
-# SqlPython V1.6.8
+# SqlPython V1.6.7
 # Author: Luca.Canali@cern.ch, Apr 2006
 # Rev 2-Sep-09
 
@@ -8,12 +8,20 @@
 # Best used with the companion modules sqlpyPlus and mysqlpy 
 # See also http://twiki.cern.ch/twiki/bin/view/PSSGroup/SqlPython
 
-import cmd2,getpass,binascii,cx_Oracle,re,os,platform
-import sqlalchemy, pyparsing
-__version__ = '1.6.8'    
+import cmd2,getpass,binascii,re,os,platform
+import pyparsing, connections
+__version__ = '1.7.0'
+try:
+    import cx_Oracle
+except ImportError:
+    cx_Oracle = None
+try:
+    import psycopg2
+except ImportError:
+    psycopg2 = None
 
 class Parser(object):
-    comment_def = "--" + ~ ('-' + pyparsing.CaselessKeyword('begin')) + pyparsing.ZeroOrMore(pyparsing.CharsNotIn("\n"))    
+    comment_def = "--" + pyparsing.NotAny('-' + pyparsing.CaselessKeyword('begin')) + pyparsing.ZeroOrMore(pyparsing.CharsNotIn("\n"))    
     def __init__(self, scanner, retainSeparator=True):
         self.scanner = scanner
         self.scanner.ignore(pyparsing.sglQuotedString)
@@ -41,165 +49,135 @@ class sqlpython(cmd2.Cmd):
 
     def __init__(self):
         cmd2.Cmd.__init__(self)
-        self.no_connection()
+        self.no_instance()
         self.maxfetch = 1000
         self.terminator = ';'
         self.timeout = 30
         self.commit_on_exit = True
-        self.connections = {}
+        self.instances = {}
         
-    def no_connection(self):
+    def no_instance(self):
         self.prompt = 'SQL.No_Connection> '
         self.curs = None
-        self.conn = None
-        self.connection_number = None
+        self.current_instance = None
+        self.instance_number = None
         
-    def make_connection_current(self, connection_number):
-        self.conn = self.connections[connection_number]['conn']
-        self.prompt = self.connections[connection_number]['prompt']
-        self.rdbms = self.connections[connection_number]['rdbms']
-        self.connection_number = connection_number
-        self.curs = self.conn.cursor()
+    def make_instance_current(self, instance_number):
+        db_instance = self.instances[instance_number]
+        self.prompt = db_instance.prompt
+        self.rdbms = db_instance.rdbms
+        self.instance_number = instance_number
+        self.curs = db_instance.connection.cursor()
+        self.current_instance = db_instance
             
-    def successful_connection_to_number(self, arg):
+    def list_instances(self):
+        self.stdout.write('Existing connections:\n')
+        self.stdout.write('\n'.join('%s (%s)' % (v.prompt, v.rdbms) 
+                                    for (k,v) in sorted(self.instances.items())) + '\n')
+        
+    def disconnect(self, arg):
         try:
-            connection_number = int(arg)
+            instance_number = int(arg)
+            instance = self.instances[instance_number]
+        except (ValueError, KeyError):
+            self.list_instances()
+            return
+        if self.commit_on_exit:
+            instance.connection.commit()
+        self.instances.pop(instance_number)
+        if instance_number == self.instance_number:
+            self.no_instance()
+            
+    def closeall(self):
+        for instance_number in self.instances.keys():
+            self.disconnect(instance_number)
+        self.curs = None
+        self.no_instance()        
+            
+    legal_sql_word = pyparsing.Word(pyparsing.alphanums + '_$#')
+          
+    def successfully_connect_to_number(self, arg):
+        try:
+            instance_number = int(arg)
         except ValueError:            
             return False
-        self.make_connection_current(connection_number)
+        try:
+            self.make_instance_current(instance_number)
+        except IndexError:
+            self.list_instances()
+            return False
         if (self.rdbms == 'oracle') and self.serveroutput:
             self.curs.callproc('dbms_output.enable', [])           
         return True
 
-    def list_connections(self):
-        self.stdout.write('Existing connections:\n')
-        self.stdout.write('\n'.join('%s (%s)' % (v['prompt'], v['rdbms']) 
-                                    for (k,v) in sorted(self.connections.items())) + '\n')
-        
-    def disconnect(self, arg):
-        try:
-            connection_number = int(arg)
-            connection = self.connections[connection_number]
-        except (ValueError, KeyError):
-            self.list_connections()
-            return
-        if self.commit_on_exit:
-            connection['conn'].commit()
-        self.connections.pop(connection_number)
-        if connection_number == self.connection_number:
-            self.no_connection()
-            
-    def closeall(self):
-        for connection_number in self.connections.keys():
-            self.disconnect(connection_number)
-        self.curs = None
-        self.no_connection()        
-            
-    def url_connect(self, arg):
-        eng = sqlalchemy.create_engine(arg) 
-        self.conn = eng.connect().connection
-        conn  = {'conn': self.conn, 'prompt': self.prompt, 'dbname': eng.url.database,
-                 'rdbms': eng.url.drivername, 'user': eng.url.username or '', 
-                 'eng': eng}
-        return conn
-    def ora_connect(self, arg):
-        modeval = 0
-        oraserv = None
-        for modere, modevalue in self.connection_modes.items():
-            if modere.search(arg):
-                arg = modere.sub('', arg)
-                modeval = modevalue
-        try:
-            orauser, oraserv = arg.split('@')
-        except ValueError:
-            try:
-                oraserv = os.environ['ORACLE_SID']
-            except KeyError:
-                self.perror('instance not specified and environment variable ORACLE_SID not set')
-                return
-            orauser = arg
-        sid = oraserv
-        try:
-            host, sid = oraserv.split('/')
-            try:
-                host, port = host.split(':')
-                port = int(port)
-            except ValueError:
-                port = 1521
-            oraserv = cx_Oracle.makedsn(host, port, sid)
-        except ValueError:
-            pass
-        try:
-            orauser, orapass = orauser.split('/')
-        except ValueError:
-            orapass = getpass.getpass('Password: ')
-        if orauser.upper() == 'SYS' and not modeval:
-            self.pfeedback('Privilege not specified for SYS, assuming SYSOPER')
-            modeval = cx_Oracle.SYSOPER
-        result = self.url_connect('oracle://%s:%s@%s/?mode=%d' % (orauser, orapass, oraserv, modeval))
-        result['dbname'] = oraserv
-        return result
-    
-    connection_modes = {re.compile(' AS SYSDBA', re.IGNORECASE): cx_Oracle.SYSDBA, 
-                        re.compile(' AS SYSOPER', re.IGNORECASE): cx_Oracle.SYSOPER}
-    @cmd2.options([cmd2.make_option('-a', '--add', action='store_true', 
+    @cmd2.options([cmd2.make_option('-a', '--add', action='store_true',
                                     help='add connection (keep current connection)'),
-                   cmd2.make_option('-c', '--close', action='store_true', 
+                   cmd2.make_option('-c', '--close', action='store_true',
                                     help='close connection {N} (or current)'),
-                   cmd2.make_option('-C', '--closeall', action='store_true', 
-                                    help='close all connections'),])
+                   cmd2.make_option('-C', '--closeall', action='store_true',
+                                    help='close all connections'),
+                   cmd2.make_option('--postgres', action='store_true', help='Connect to postgreSQL: `sqlpython --postgres [DBNAME [USERNAME]]`'),
+                   cmd2.make_option('--oracle', action='store_true', help='Connect to an Oracle database'),
+                   cmd2.make_option('--mysql', action='store_true', help='Connect to a MySQL database'),
+                   cmd2.make_option('-H', '--hostname', type='string',
+                                    help='Machine where database is hosted'),
+                   cmd2.make_option('-p', '--port', type='int',
+                                    help='Port to connect to'),
+                   cmd2.make_option('--password', type='string',
+                                    help='Password'),
+                   cmd2.make_option('-d', '--database', type='string',
+                                    help='Database name to connect to'),
+                   cmd2.make_option('-U', '--username', type='string',
+                                    help='Database user name to connect as')])
     def do_connect(self, arg, opts):
+ 
         '''Opens the DB connection'''
         if opts.closeall:
             self.closeall()
-            return
+            return 
         if opts.close:
             if not arg:
-                arg = self.connection_number
+                arg = self.instance_number
             self.disconnect(arg)
+            return 
+        if (not arg) and (not opts.postgres) and (not opts.mysql):
+            self.list_instances()
+            return 
+        if self.successfully_connect_to_number(arg):
             return
-        if not arg:
-            self.list_connections()
-            return
-        try:
-            if self.successful_connection_to_number(arg):
-                return
-        except IndexError:
-            self.list_connections()
-            return
-        try:
-            connect_info = self.url_connect(arg)
-        except sqlalchemy.exc.ArgumentError, e:
-            connect_info = self.ora_connect(arg)
-        if opts.add or (self.connection_number is None):
+        
+        db_instance = connections.DatabaseInstance(arg, opts, default_rdbms = self.default_rdbms)
+        if opts.add or (self.instance_number is None):
             try:
-                self.connection_number = max(self.connections.keys()) + 1
+                self.instance_number = max(self.instances.keys()) + 1
             except ValueError:
-                self.connection_number = 0
-        connect_info['prompt'] = '%d:%s@%s> ' % (self.connection_number, connect_info['user'], connect_info['dbname'])
-        self.connections[self.connection_number] = connect_info
-        self.make_connection_current(self.connection_number)
-        self.curs = self.conn.cursor()
+                self.instance_number = 0            
+        db_instance.set_instance_number(self.instance_number)
+        self.instances[self.instance_number] = db_instance
+        self.make_instance_current(self.instance_number)        
         if (self.rdbms == 'oracle') and self.serveroutput:
-            self.curs.callproc('dbms_output.enable', [])
-        if (self.rdbms == 'mysql'):
-            self.curs.execute('SET SQL_MODE=ANSI')
+            self.current_instance.connection.cursor().callproc('dbms_output.enable',[])
+            #self.curs.callproc('dbms_output.enable', [])        
+    
+    def do_pickle(self, arg):
+        self.current_instance.pickle()
+        
     def postparsing_precmd(self, statement):
         stop = 0
-        self.saved_connection_number = None
-        if statement.parsed.connection_number:
-            saved_connection_number = self.connection_number
+        self.saved_instance_number = None
+        if statement.parsed.instance_number:
+            saved_instance_number = self.instance_number
             try:
-                if self.successful_connection_to_number(statement.parsed.connection_number):
+                if self.successfully_connect_to_number(statement.parsed.instance_number):
                     if statement.parsed.command:
-                        self.saved_connection_number = saved_connection_number
+                        self.saved_instance_number = saved_instance_number
             except KeyError:
-                self.list_connections()
-                raise KeyError, 'No connection #%s' % statement.parsed.connection_number
+                self.list_instances()
+                raise KeyError, 'No connection #%s' % statement.parsed.instance_number
         return stop, statement           
     def postparsing_postcmd(self, stop):
-        if self.saved_connection_number is not None:
-            self.successful_connection_to_number(self.saved_connection_number)
+        if self.saved_instance_number is not None:
+            self.successfully_connect_to_number(self.saved_instance_number)
         return stop
                 
     do_host = cmd2.Cmd.do_shell
@@ -272,10 +250,10 @@ class sqlpython(cmd2.Cmd):
     terminatorSearchString = '|'.join('\\' + d.split()[0] for d in do_terminators.__doc__.splitlines())
         
     bindScanner = {'oracle': Parser(pyparsing.Literal(':') + pyparsing.Word( pyparsing.alphanums + "_$#" )),
-                   'postgres': Parser(pyparsing.Literal('%(') + 
-                                      pyparsing.Word(pyparsing.alphanums + "_$#") + ')s')}
+                   'postgres': Parser(pyparsing.Literal('%(') + legal_sql_word + ')s')}
     def findBinds(self, target, givenBindVars = {}):
         result = givenBindVars
+        #TODO: A consistent bind style?  As a setting, perhaps?
         if self.rdbms in self.bindScanner:
             for finding, startat, endat in self.bindScanner[self.rdbms].scanner.scanString(target):
                 varname = finding[1]
@@ -283,7 +261,7 @@ class sqlpython(cmd2.Cmd):
                     result[varname] = self.binds[varname]
                 except KeyError:
                     if not givenBindVars.has_key(varname):
-                        print 'Bind variable %s not defined.' % (varname)
+                        raise KeyError, 'Bind variable "%s" not defined.' % (varname)
         return result
 
     def default(self, arg):
@@ -295,7 +273,8 @@ class sqlpython(cmd2.Cmd):
             command = '%s %s'    
         if self.rdbms == 'oracle':
             current_time = self.current_database_time()
-        self.curs.execute(command % (arg.parsed.command, arg.parsed.args), self.varsUsed)
+        commandstring = command % (arg.parsed.command, arg.parsed.args)
+        self.curs.execute(commandstring, self.varsUsed)
         executionmessage = '\nExecuted%s\n' % ((self.curs.rowcount > 0) and ' (%d rows)' % self.curs.rowcount or '')
         if self.rdbms == 'oracle':
             self._show_errors(all_users=True, limit=1, mintime=current_time)
@@ -311,37 +290,26 @@ class sqlpython(cmd2.Cmd):
         return cmd2.Cmd.do_quit(self, None)
     do_exit = do_quit
     do_q = do_quit
-    colorcodes = {'bold':{True:'\x1b[1m',False:'\x1b[22m'},
-                  'red':{True:'\x1b[36m',False:'\x1b[39m'},
-                  'cyan':{True:'\x1b[31m',False:'\x1b[39m'},
-                  'underline':{True:'\x1b[4m',False:'\x1b[24m'}}
-    colors = (platform.system() != 'Windows')
     def colorize(self, val, color):
-        if self.colors and (self.stdout == self.initial_stdout):
-            if color not in self.colorcodes:
-                if (color % 2):
-                    color = 'red'
-                else:
-                    color = 'cyan'
-            return self.colorcodes[color][True] + val + self.colorcodes[color][False]        
-        return val
-    def pmatrix(self,rows,desc,maxlen=30,heading=True,restructuredtext=False):
+        if color not in self.colorcodes:
+            if (color % 2):
+                color = 'red'
+            else:
+                color = 'cyan'
+        return cmd2.Cmd.colorize(self, val, color)
+    def pmatrix(self,rows,maxlen=30,heading=True,restructuredtext=False):
         '''prints a matrix, used by sqlpython to print queries' result sets'''
-        names = []
-        maxen = []
+        names = self.colnames
+        maxen = [len(n) for n in self.colnames]
         toprint = []
-        for d in desc:
-            n = d[0]
-            names.append(n)      # list col names
-            maxen.append(len(n)) # col length
-        rcols = range(len(desc))
+        rcols = range(len(self.colnames))
         rrows = range(len(rows))
         for i in rrows:          # loops for all rows
             rowsi = map(str, rows[i]) # current row to process
             split = []                # service var is row split is needed
             mustsplit = 0             # flag 
             for j in rcols:
-                if str(desc[j][1]) == "<type 'cx_Oracle.BINARY'>":  # handles RAW columns
+                if str(self.coltypes[j]) == "<type 'cx_Oracle.BINARY'>":  # handles RAW columns
                     rowsi[j] = binascii.b2a_hex(rowsi[j])
                 maxen[j] = max(maxen[j], len(rowsi[j]))    # computes max field length
                 if maxen[j] <= maxlen:
@@ -363,7 +331,7 @@ class sqlpython(cmd2.Cmd):
             rrows2 = range(len(toprint))
             for j in rrows2:
                 val = toprint[j][i]
-                if str(desc[i][1]) == "<type 'cx_Oracle.NUMBER'>":  # right align numbers
+                if str(self.coltypes[i]) == "<type 'cx_Oracle.NUMBER'>":  # right align numbers - but must generalize!
                     toprint[j][i] = (" " * (maxcol-len(val))) + val
                 else:
                     toprint[j][i] = val + (" " * (maxcol-len(val)))
